#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <string>
#include <math.h>
#include <cuda_runtime.h>
#include <device_launch_parameters.h>

#define gpuErrChk(ans) { gpuAssert((ans), __FILE__, __LINE__); }
inline void gpuAssert(cudaError_t code, const char* file, int line,
    bool abort = true)
{
    if (code != cudaSuccess) {
        fprintf(stderr, "GPUassert: %s %s %d\n",
            cudaGetErrorString(code), file, line);
        exit(code);
    }
}

__host__ void cpuSum(const int* const input, int* output, int n) {
    int i;

    for (i = 0; i < n; i++) {
        *output += input[i];
    }
}

__global__ void cudaSum_Naive(const int* const input, int* output, int n) {
    int partial_sum = 0.0;

    int idx = threadIdx.x * n + blockIdx.x * blockDim.x * n;
    int end = idx + n;
    while (idx < end) {
        partial_sum += input[idx];
        idx++;
    }

    atomicAdd(output, partial_sum);
}

__global__ void cudaSum_Binary(const int* const input, int* output) {
    extern __shared__ int shmem[];

    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    shmem[threadIdx.x] = input[idx];
    __syncthreads();

    for (int i = 1; i < blockDim.x; i *= 2) {
        if (threadIdx.x % (2 * i) == 0) {
            shmem[threadIdx.x] += shmem[threadIdx.x + i];
        }
        __syncthreads();
    }

    if (threadIdx.x == 0) {
        output[blockIdx.x] = shmem[0];
    }
}

__global__ void cudaSum_NonDivergent(const int* const input, int* output) {
    extern __shared__ int shmem[];

    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    shmem[threadIdx.x] = input[idx];
    __syncthreads();

    for (int i = 1; i < blockDim.x; i *= 2) {
        int j = 2 * i * threadIdx.x;
        if (j < blockDim.x) {
            shmem[j] += shmem[i + j];
        }
        __syncthreads();
    }

    if (threadIdx.x == 0) {
        output[blockIdx.x] = shmem[0];
    }

}

__global__ void cudaSum_seqAddressing(const int* const input, int* output) {
    extern __shared__ int shmem[];

    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    shmem[threadIdx.x] = input[idx];
    __syncthreads();

    for (int i = blockDim.x / 2; i > 0; i = i / 2) {
        if (threadIdx.x < i) {
            shmem[threadIdx.x] += shmem[threadIdx.x + i];
        }
        __syncthreads();
    }

    if (threadIdx.x == 0) {
        output[blockIdx.x] = shmem[0];
    }
}




/*
 * Fills fill with random numbers is [0, 1]. Size is number of elements to
 * assign.
 */
void randomFill(int* fill, int size) {
    for (int i = 0; i < size; i++) {
        int r = rand();
        fill[i] = r;
    }
}

void randomFillInt(int* fill, int size) {
    for (int i = 0; i < size; i++) {
        int r = rand();
        fill[i] = r;
    }
}

/*
 * Fills fill with random numbers is [0, 1]. Size is number of elements to
 * assign.
 */
void onesFill(int* fill, int size) {
    for (int i = 0; i < size; i++) {
        fill[i] = 1.0;
    }
}

void outputArray(int* arr, int size) {
    for (int i = 0; i < size; i++) {
        printf("%d: %d\n", i, arr[i]);
    }
}

int isEqual(int* a, int* b) {
    if (*a == *b) {
        return 1;
    }
    return 0;
}

int main() {
    cudaEvent_t start;
    cudaEvent_t stop;

#define START_TIMER() {                                                        \
            gpuErrChk(cudaEventCreate(&start));                                \
            gpuErrChk(cudaEventCreate(&stop));                                 \
            gpuErrChk(cudaEventRecord(start));                                 \
        }

#define STOP_RECORD_TIMER(name) {                                              \
            gpuErrChk(cudaEventRecord(stop));                                  \
            gpuErrChk(cudaEventSynchronize(stop));                             \
            gpuErrChk(cudaEventElapsedTime(&name, start, stop));               \
            gpuErrChk(cudaEventDestroy(start));                                \
            gpuErrChk(cudaEventDestroy(stop));                                 \
        }
    printf("n,cpu,naive_gpu,bin_gpu,nondiv_gpu,seqaddr_gpu\n");
    for (int t = 10; t < 22; t++) {
        // Initialize timers
        float cpu_ms = -1;
        float naive_gpu_ms = -1;
        float binary_gpu_ms = -1;
        float nondivergent_gpu_ms = -1;
        float sequential_gpu_ms = -1;

        int n = 1<<t;
        int numberOfInputs = 2;
        srand(2021);

        // allocate host memory
        int* input = new int[n];
        int* output = (int*)malloc(sizeof(int));
        int* gpu_output = (int*)malloc(sizeof(int));

        *output = 0;
        *gpu_output = 0;

        // allocate device memory
        int* d_input;
        int* d_output;

        gpuErrChk(cudaMalloc(&d_input, n * sizeof(int)));
        gpuErrChk(cudaMalloc(&d_output, sizeof(int)));

        // randomly fill input
        randomFill(input, n);

        // copy input to GPU
        gpuErrChk(cudaMemcpy(d_input, input, n * sizeof(int),
            cudaMemcpyHostToDevice));

        // CPU implementation
        START_TIMER();
        cpuSum(input, output, n);
        STOP_RECORD_TIMER(cpu_ms);
        // printf("Size %d CPU: %f ms\n", n, cpu_ms);

        // GPU implementation
        int blockSize = n / 256;

        // Naive
        START_TIMER();
        cudaSum_Naive << <blockSize, ((n / blockSize) / numberOfInputs) >> > (d_input, d_output, numberOfInputs);
        gpuErrChk(cudaMemcpy(gpu_output, d_output, sizeof(int),
            cudaMemcpyDeviceToHost));
        STOP_RECORD_TIMER(naive_gpu_ms);
        // printf("Size %d naive GPU: %f ms\n", n, naive_gpu_ms);
        assert(isEqual(output, gpu_output));

        // reset gpu output variable
        memset(gpu_output, 0, sizeof(int));
        cudaMemset(d_output, 0, sizeof(int));
        *gpu_output = 0;

        // declare new array output variable
        int* gpu_array_output = new int[blockSize];
        int* d_array_output;

        // Binary Tree
        gpuErrChk(cudaMalloc(&d_array_output, blockSize * sizeof(int)));
        START_TIMER();
        cudaSum_Binary << <blockSize, 256, 256 * sizeof(int) >> > (d_input, d_array_output);
        gpuErrChk(cudaMemcpy(gpu_array_output, d_array_output, blockSize * sizeof(int),
            cudaMemcpyDeviceToHost));
        cpuSum(gpu_array_output, gpu_output, blockSize);
        STOP_RECORD_TIMER(binary_gpu_ms);
        // printf("Size %d Binary Tree GPU: %f ms\n", n, binary_gpu_ms);
        assert(isEqual(output, gpu_output));

        // reset gpu output variable
        memset(gpu_array_output, 0, blockSize * sizeof(int));
        cudaMemset(d_array_output, 0, blockSize * sizeof(int));
        *gpu_output = 0;

        // Non-Divergent
        START_TIMER();
        cudaSum_NonDivergent << <blockSize, 256, 256 * sizeof(int) >> > (d_input, d_array_output);
        gpuErrChk(cudaMemcpy(gpu_array_output, d_array_output, blockSize * sizeof(int),
            cudaMemcpyDeviceToHost));
        cpuSum(gpu_array_output, gpu_output, blockSize);
        STOP_RECORD_TIMER(nondivergent_gpu_ms);
        // printf("Size %d Non-Divergent GPU: %f ms\n", n, nondivergent_gpu_ms);
        assert(isEqual(output, gpu_output));

        // reset gpu output variable
        memset(gpu_array_output, 0, blockSize * sizeof(int));
        cudaMemset(d_array_output, 0, blockSize * sizeof(int));
        *gpu_output = 0;

        // Sequential Addressing
        START_TIMER();
        cudaSum_seqAddressing << <blockSize, 256, 256 * sizeof(int) >> > (d_input, d_array_output);
        gpuErrChk(cudaMemcpy(gpu_array_output, d_array_output, blockSize * sizeof(int),
            cudaMemcpyDeviceToHost));
        cpuSum(gpu_array_output, gpu_output, blockSize);
        STOP_RECORD_TIMER(sequential_gpu_ms);
        // printf("Size %d Sequential GPU: %f ms\n", n, sequential_gpu_ms);
        assert(isEqual(output, gpu_output));

        // reset gpu output variable
        memset(gpu_array_output, 0, blockSize * sizeof(int));
        cudaMemset(d_array_output, 0, blockSize * sizeof(int));
        *gpu_output = 0;

        printf("%d,%f,%f,%f,%f,%f\n", n, cpu_ms, naive_gpu_ms, binary_gpu_ms, nondivergent_gpu_ms, sequential_gpu_ms);

        // free CPU memory
        delete[] input;
        delete output;
        delete gpu_output;
        delete[] gpu_array_output;

        // free GPU memory
        gpuErrChk(cudaFree(d_input));
        gpuErrChk(cudaFree(d_output));
        gpuErrChk(cudaFree(d_array_output));
    }

    return 0;
}
